unit WaitThread;

interface

uses
  SysUtils, Classes, Windows, FileCtrl;

const
  FSM_DIR_NOEXIST = 1;

type
  EFileSysMonitor = class(Exception);

  TFileSysNotify = (FILE_NAME,
                    DIR_NAME,
                    ATTRIBUTES,
                    SIZE,
                    LAST_WRITE,
                    SECURITY);
  TNotifyFilter = set of TFileSysNotify;

  TFileSysChange = procedure(Sender: TObject; Directory: string) of object;

  TFileWaitThread = class(TThread)
  private
    FNotifyHandle    : THandle;
    lpPathName       : PChar;
    bWatchSubtree    : boolean;
    dwNotifyFilter   : DWord;
    FTimeOut         : DWord;
    FDoFileSysChange : TFileSysChange;
    procedure HandleDirChange;
    procedure ReportError;
  protected
    procedure Execute; override;
    destructor Destroy; override;
  public
    constructor Create(Directory: string;
                       WatchSubtree: boolean;
                       NotifyFilter: DWord;
                       CallBack: TFileSysChange);  //redefining the create, so I do not override here
  end;

  TFileSysMonitor = class(TComponent)
  private
    FWaitThread : TFileWaitThread;
    FDirectory : string;
    FWatchSubtree : boolean;
    FNotifyFilter : TNotifyFilter;
    FOnFileSysChange : TFileSysChange;
    FActive : boolean;
    procedure RaiseError(errorcode: integer);
    function GetActive: boolean;
    procedure SetActive(Value: boolean);
    procedure SetDirectory(Value: string);
    procedure SetWatchSubtree(Value: boolean);
    procedure SetNotifyFilter(Value: TNotifyFilter);
    function GetFilterAsWord: Word;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property MonitoredDirectory : string
                                  read FDirectory
                                  write SetDirectory;
    property WatchSubtree : boolean
                            read FWatchSubtree
                            write SetWatchSubtree
                            default False;
    property NotifyFilter : TNotifyFilter
                            read FNotifyFilter
                            write SetNotifyFilter
                            default [FILE_NAME];
    property OnFileSysChange : TFileSysChange
                               read FOnFileSysChange
                               write FOnFileSysChange;
    property Active : boolean
                      read GetActive
                      write SetActive
                      default False;
  end;

procedure Register;

implementation

function TFileSysMonitor.GetActive: boolean;
begin
  if FWaitThread = nil then
    FActive := False;
  Result := FActive;
end;

procedure TFileSysMonitor.SetActive(Value: boolean);
begin
  if Value <> FActive then
    begin
    case Value of
      True  : begin
              FWaitThread := TFileWaitThread.Create(FDirectory,
                                                    FWatchSubtree,
                                                    GetFilterAsWord,
                                                    FOnFileSysChange);
              FWaitThread.Resume;
              end;
      False : begin
              if FWaitThread <> nil then
                FWaitThread.Terminate;
              FWaitThread := nil;
              end;
      end; //case
    FActive := Value;
    end;
end;

procedure TFileSysMonitor.SetDirectory(Value: string);
begin
  if Value <> FDirectory then
    begin
    if DirectoryExists(Value) then
      begin
      FDirectory := Value;
      if Active then   //Reset active monitor with new path
        begin
        Active := False;
        Active := True;
        end;
      end
    else
      RaiseError(FSM_DIR_NOEXIST);  //No such directory!!
    end;
end;

procedure TFileSysMonitor.SetWatchSubtree(Value: boolean);
begin
  if Value <> FWatchSubtree then
    begin
    FWatchSubtree := Value;
    if Active then   //Reset active monitor with new settings
      begin
      Active := False;
      Active := True;
      end;
    end;
end;

procedure TFileSysMonitor.SetNotifyFilter(Value: TNotifyFilter);
begin
  if Value <> FNotifyFilter then
    begin
    FNotifyFilter := Value;
    if Active then   //Reset active monitor with new filter
      begin
      Active := False;
      Active := True;
      end;
    end;
end;

function TFileSysMonitor.GetFilterAsWord: Word;
var
  wTemp : Word;
begin
  wTemp := 0;
  if FILE_NAME in FNotifyFilter then
    wTemp := wTemp + FILE_NOTIFY_CHANGE_FILE_NAME;
  if DIR_NAME in FNotifyFilter then
    wTemp := wTemp + FILE_NOTIFY_CHANGE_DIR_NAME;
  if ATTRIBUTES in FNotifyFilter then
    wTemp := wTemp + FILE_NOTIFY_CHANGE_ATTRIBUTES;
  if SIZE in FNotifyFilter then
    wTemp := wTemp + FILE_NOTIFY_CHANGE_SIZE;
  if LAST_WRITE in FNotifyFilter then
    wTemp := wTemp + FILE_NOTIFY_CHANGE_LAST_WRITE;
  if SECURITY in FNotifyFilter then
    wTemp := wTemp + FILE_NOTIFY_CHANGE_SECURITY;

  Result := wTemp;
end;

constructor TFileSysMonitor.Create(AOwner: TComponent);
var
  lpDirectory : PChar;
  uiSize      : UInt;
begin
  inherited Create(AOwner);

  //Use Windows dir as default for monitoring (this guarantees a valid dir name)
  lpDirectory := StrAlloc(MAX_PATH + 1);
  uiSize := GetWindowsDirectory(lpDirectory, MAX_PATH + 1);
  lpDirectory[MAX_PATH] := Char(0);  //just in case...
  FDirectory := StrPas(lpDirectory);
  StrDispose(lpDirectory);

  FWatchSubtree := False;
  FNotifyFilter := [FILE_NAME];
  FActive := False;
end;

procedure TFileSysMonitor.RaiseError(errorcode: integer);
Var
  StError:String;
Begin
  case errorcode of
    FSM_DIR_NOEXIST  : StError:='Directory does not exist!';
    else StError:='Unknown error in TFileSysMonitor component!';
    end;  //case
  Raise EFileSysMonitor.Create(StError);
End;

constructor TFileWaitThread.Create(Directory: string;
                                   WatchSubtree: boolean;
                                   NotifyFilter: DWord;
                                   CallBack: TFileSysChange);
const
  IsSuspended = True;
begin
  inherited Create(IsSuspended);  //Start up thread in paused state
  bWatchSubtree := WatchSubtree;
  dwNotifyFilter := NotifyFilter;
  FTimeOut := 100;
  lpPathName := StrAlloc(Length(Directory) + 1);
  lpPathName := StrPCopy(lpPathName, Directory);
  FDoFileSysChange := CallBack;
end;

destructor TFileWaitThread.Destroy;
begin
  StrDispose(lpPathName);
  inherited Destroy;
end;

procedure TFileWaitThread.Execute;
var
  WaitResponse : DWord;
  Finished : Boolean;
begin
  if not Terminated then
    FNotifyHandle := FindFirstChangeNotification(lpPathName,
                                                 bWatchSubtree,
                                                 dwNotifyFilter);
  if not Terminated then
    begin
    if FNotifyHandle = INVALID_HANDLE_VALUE then
      //Error occurred setting up notification
      Synchronize(ReportError)
      else
        begin
        if not Terminated then  //Check before entering loop
          begin
          //Wait for a notification
          Finished := False;
          while ((not Terminated) and (not Finished)) do
            begin
            WaitResponse := WaitForSingleObject( FNotifyHandle, FTimeOut );
            case WaitResponse of
              WAIT_FAILED   : begin
                              Finished := True;
                              Synchronize(ReportError);
                              end;
              WAIT_OBJECT_0 : Synchronize(HandleDirChange); //A Change was detected in file system
              end; //case
            if not FindNextChangeNotification(FNotifyHandle) then
              Synchronize(ReportError);
            end;
          end;
        FindCloseChangeNotification(FNotifyHandle);  //stop monitoring the file system
        end;
    end;
end;

procedure TFileWaitThread.ReportError;
var
  stError : string;
begin
  stError := 'TFileWaitThread Error, code ' + IntToStr(GetLastError());
  Raise EFileSysMonitor.Create(StError);
end;

procedure TFileWaitThread.HandleDirChange;
var
  strTemp : string;
begin
  if Assigned(FDoFileSysChange) then
    begin
    strTemp := StrPas(lpPathName);
    FDoFileSysChange(Self, strTemp);
    end;
end;

procedure Register;
begin
  RegisterComponents('The Good Package', [TFileSysMonitor]);
end;

end.
